from .mixins import DxMixin
import rest_framework.viewsets
from django.db.models import Count
from rest_framework.response import Response
from .pagination import TakeSkipPagination
from .filters import DxFilterBackend


class DxModelViewSet(rest_framework.viewsets.ModelViewSet, DxMixin):
    pagination_class = TakeSkipPagination
    filter_backends = [DxFilterBackend, *rest_framework.viewsets.ModelViewSet.filter_backends]

    def list(self, request, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())
        group = self.get_param_from_request(request, "group")
        if group is None:
            return self._not_grouped_list(queryset)
        else:
            return self._grouped_list(group, queryset, request)

    def _grouped_list(self, group, queryset, request):
        require_group_count = self.get_param_from_request(request, "requireGroupCount")
        require_total_count = self.get_param_from_request(request, "requireTotalCount")
        if group[0]["isExpanded"]:
            return Response()
        else:
            group_field_name = group[0]["selector"].replace(".", "__")
            ordering = self.get_ordering(group)
            group_queryset = queryset.values(group_field_name).annotate(count=Count("pk")).order_by(
                *ordering).distinct()
            page = self.paginate_queryset(group_queryset)
            res_dict = {}
            if require_total_count is None and require_group_count is None:
                res_dict["totalCount"] = group_queryset.count()
            else:
                if require_group_count:
                    res_dict["groupCount"] = group_queryset.count()
                if require_total_count:
                    res_dict["totalCount"] = queryset.count()
            if page is not None:
                res_dict["data"] = [{"key": x[group_field_name], "items": None, "count": x["count"]} for x in page]
                return Response(res_dict)
            else:
                res_dict["data"] = [{"key": x[group_field_name], "items": None, "count": x["count"]} for x in
                                    group_queryset]
            return Response(res_dict)

    def _not_grouped_list(self, queryset):
        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)
        serializer = self.get_serializer(queryset, many=True)
        return Response({"data": serializer.data})



from rest_framework.pagination import LimitOffsetPagination, _positive_int
from rest_framework.response import Response
from collections import OrderedDict


class TakeSkipPagination(LimitOffsetPagination):
    limit_query_param = 'take'
    offset_query_param = 'skip'

    def get_limit(self, request):
        if self.limit_query_param:
            try:
                return _positive_int(
                    request.query_params[self.limit_query_param] if self.limit_query_param in request.query_params
                    else request.data[self.limit_query_param],
                    strict=True,
                    cutoff=self.max_limit
                )
            except (KeyError, ValueError):
                pass

        return self.default_limit

    def get_offset(self, request):
        try:
            return _positive_int(
                request.query_params[self.offset_query_param] if self.offset_query_param in request.query_params
                else request.data[self.offset_query_param],
            )
        except (KeyError, ValueError):
            return 0

    def paginate_queryset(self, queryset, request, view=None):
        self.count = self.get_count(queryset)
        self.limit = self.get_limit(request)
        self.offset = self.get_offset(request)
        if self.limit is None and self.offset is None:
            return None
        self.request = request

        if self.limit is not None and self.count > self.limit and self.template is not None:
            self.display_page_controls = True

        if self.count == 0 or self.offset > self.count:
            return []

        if self.limit is None:
            return list(queryset[self.offset:])
        return list(queryset[self.offset:self.offset + self.limit])

    def get_paginated_response(self, data):
        return Response(OrderedDict([
            ('totalCount', self.count),
            ('data', data)
        ]))



from rest_framework import filters
from django.db.models import Q
from .mixins import DxMixin


class DxFilterBackend(filters.BaseFilterBackend, DxMixin):

    def __is_leaf(self, dx_filter):
        for elem in dx_filter:
            if isinstance(elem, list):
                return False
        return True

    def __to_django_operator(self, operator):
        if operator == "=":
            return ""
        elif operator == ">":
            return "__gt"
        elif operator == "<":
            return "__lt"
        elif operator == ">=":
            return "__gte"
        elif operator == "<=":
            return "__lte"
        else:
            return "__" + operator

    def __leaf_node_to_q(self, leaf):
        field_name = leaf[0].replace(".", "__")
        if len(leaf) == 2 and leaf[1] == "=":
            kwargs = {}
            kwargs[field_name + "__isnull"] = True
            return Q(**kwargs)
        elif leaf[1] == "<>":
            kwargs = {}
            kwargs[field_name] = leaf[2]
            return ~Q(**kwargs)
        elif leaf[1] == "notcontains":
            kwargs = {}
            kwargs[field_name + "__contains"] = leaf[2]
            return ~Q(**kwargs)
        else:
            kwargs = {}
            kwargs[field_name + self.__to_django_operator(leaf[1])] = leaf[2]
            return Q(**kwargs)

    def __generate_q_expr(self, dx_filter):
        if dx_filter is None or not dx_filter:
            return None
        if self.__is_leaf(dx_filter):
            return self.__leaf_node_to_q(dx_filter)
        else:
            q_elems = []

            for elem in dx_filter:
                if isinstance(elem, list):
                    q_elems.append(self.__generate_q_expr(elem))
                elif elem in ["and", "or"]:
                    q_elems.append(elem)
                else:
                    raise Exception("Невозможно применить данный поиск")
            q_expr = q_elems[0]
            for num_pair in range(0, len(q_elems) // 2):
                oper = q_elems[num_pair * 2 + 1]
                if oper == "and":
                    q_expr = q_expr & q_elems[(num_pair + 1) * 2]
                else:
                    q_expr = q_expr | q_elems[(num_pair + 1) * 2]
            return q_expr

    def filter_queryset(self, request, queryset, view):
        res_queryset = queryset
        filter = self.get_param_from_request(request, "filter")
        if filter:
            q_expr = self.__generate_q_expr(filter)
            if q_expr is not None:
                res_queryset = res_queryset.filter(q_expr)
        sort = self.get_param_from_request(request, "sort")
        if sort:
            ordering = self.get_ordering(sort)
            if ordering:
                res_queryset = res_queryset.order_by(*ordering)
        return res_queryset



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='drf_dx_datagrid',
    version='0.1.2',
    packages=find_packages(),
    url='https://github.com/SergeyMatveev88/drf-dx-datagrid',
    license='MIT',
    author='SergeyMatveev88',
    author_email='dazranagon@yandex.ru',
    description='This package provides easy integration between Django REST framework and DevExtreme Data Grid. It handles grouping, paging, filtering and ordering on serverside.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=['djangorestframework'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        'Framework :: Django'
    ],
    python_requires='>=3.5',
)



import json


class DxMixin(object):
    @classmethod
    def get_param_from_request(cls, request, param_name):
        def json_loads(value, default):
            try:
                json_object = json.loads(value)
            except json.JSONDecodeError as e:
                return default
            return json_object

        if param_name + "[]" in request.query_params:
            raw_list = request.query_params.getlist(param_name + "[]")
            return [json_loads(x, x) for x in raw_list]
        if param_name in request.query_params:
            param_value = request.query_params.get(param_name)
            if param_value != '':
                return json.loads(param_value)
        return request.data.get(param_name)

    @classmethod
    def get_ordering(cls, dx_sort_list):
        if dx_sort_list is None:
            return []
        result = []
        for param in dx_sort_list:
            result.append(("-" if "desc" in param and param["desc"] else '') + param["selector"].replace(".", "__"))
        return result
